@@ -11,9 +11,6 @@ import time
 import unittest
 
 import framework
-test_host = framework.test_host
-test_port = framework.test_port
-
 
 import thor
 from thor.events import on
@@ -30,13 +27,16 @@ class TestHttpClient(framework.ClientServerTestCase):
     def create_server(self, test_host, test_port, server_side):
         class LittleRequestHandler(SocketServer.BaseRequestHandler):
             handle = server_side
-        server = LittleServer((framework.test_host, framework.test_port), LittleRequestHandler)
-        self.move_to_thread(target=server.serve_forever)
+        server = LittleServer((test_host, test_port), LittleRequestHandler)
+
+        def serve():
+            server.serve_forever(poll_interval=0.1)
+        self.move_to_thread(target=serve)
 
-        @on(self.loop)
         def stop():
             server.shutdown()
             server.server_close()
+        return stop
 
     def create_client(self, host, port, client_side):
         client = HttpClient(loop=self.loop)
@@ -91,7 +91,7 @@ class TestHttpClient(framework.ClientServerTestCase):
             def response_done(trailers):
                 self.loop.stop()
 
-            req_uri = b"http://%s:%i/basic" % (test_host, test_port)
+            req_uri = b"http://%s:%i/basic" % (framework.test_host, framework.test_port)
             exchange.request_start(b"GET", req_uri, [])
             exchange.request_done([])
 
@@ -120,7 +120,7 @@ Connection: close
             def response_done(trailers):
                 self.loop.stop()
 
-            req_uri = b"http://%s:%i/chunked_response" % (test_host, test_port)
+            req_uri = b"http://%s:%i/chunked_response" % (framework.test_host, framework.test_port)
             exchange.request_start(b"GET", req_uri, [])
             exchange.request_done([])
 
@@ -153,7 +153,7 @@ Transfer-Encoding: chunked
             def response_done(trailers):
                 self.loop.stop()
 
-            req_uri = b"http://%s:%i/chunked_request" % (test_host, test_port)
+            req_uri = b"http://%s:%i/chunked_request" % (framework.test_host, framework.test_port)
             exchange.request_start(b"POST", req_uri, [])
             exchange.request_body(req_body)
             exchange.request_body(req_body)
@@ -163,18 +163,44 @@ Transfer-Encoding: chunked
             conn.request.send(b"""\
 HTTP/1.1 200 OK
 Content-Type: text/plain
-Transfer-Encoding: chunked
+Content-Length: 5
+Connection: close
 
-5\r
-12345\r
-0\r
-\r
-""")
+12345""")
             # TODO: check server-side recv
             conn.request.close()
         self.go([server_side], [client_side])
 
 
+    def test_HEAD(self):
+        def client_side(client):
+            exchange = client.exchange()
+            self.check_exchange(exchange, {
+                'version': b"1.1",
+                'status': b"200",
+                'phrase': b'OK',
+                'body': b""
+            })
+            @on(exchange)
+            def response_done(trailers):
+                self.loop.stop()
+
+            req_uri = b"http://%s:%i/HEAD" % (framework.test_host, framework.test_port)
+            exchange.request_start(b"HEAD", req_uri, [])
+            exchange.request_done([])
+
+        def server_side(conn):
+            conn.request.send(b"""\
+HTTP/1.1 200 OK
+Content-Type: text/plain
+Content-Length: 5
+Connection: close
+
+""")
+            conn.request.close()
+        self.go([server_side], [client_side])
+
+
     def test_1xx(self):
         req_body = b"54321"
         def client_side(client):
@@ -188,16 +214,16 @@ Transfer-Encoding: chunked
             })
             @on(exchange)
             def response_nonfinal(status_code, status_phrase, headers):
+                self.assertEqual(status_code, b"110")
                 self.nonfinal_seen = True
             @on(exchange)
             def response_done(trailers):
                 self.assertTrue(self.nonfinal_seen)
                 self.loop.stop()
 
-            req_uri = b"http://%s:%i/chunked_request" % (test_host, test_port)
+            req_uri = b"http://%s:%i/chunked_request" % (framework.test_host, framework.test_port)
             exchange.request_start(b"POST", req_uri, [])
             exchange.request_body(req_body)
-            exchange.request_body(req_body)
             exchange.request_done([])
 
         def server_side(conn):
@@ -207,14 +233,10 @@ This: that
 
 HTTP/1.1 200 OK
 Content-Type: text/plain
-Transfer-Encoding: chunked
+Content-Length: 5
+Connection: close
 
-5\r
-12345\r
-0\r
-\r
-""")
-            # TODO: check server-side recv
+12345""")
             conn.request.close()
         self.go([server_side], [client_side])
 
@@ -245,7 +267,7 @@ Transfer-Encoding: chunked
             })
             exchange2.on('response_done', check_done)
 
-            req_uri = b"http://%s:%i/multiconn" % (test_host, test_port)
+            req_uri = b"http://%s:%i/multiconn" % (framework.test_host, framework.test_port)
             exchange1.request_start(b"GET", req_uri, [])
             exchange2.request_start(b"GET", req_uri, [])
             exchange1.request_done([])
@@ -259,7 +281,6 @@ Content-Length: 5
 Connection: close
 
 12345""")
-            time.sleep(1)
             conn.request.close()
         self.go([server_side], [client_side])
 
@@ -272,7 +293,7 @@ Connection: close
                 self.assertEqual(err_msg.__class__, thor.http.error.ConnectError)
                 self.loop.stop()
 
-            req_uri = b"http://%s:%i/conn_refuse_err" % (test_host, test_port)
+            req_uri = b"http://%s:%i/conn_refuse_err" % (framework.refuse_host, framework.refuse_port)
             exchange.request_start(b"GET", req_uri, [])
             exchange.request_done([])
         self.go([], [client_side])
@@ -291,8 +312,7 @@ Connection: close
         exchange.request_start(b"GET", req_uri, [])
         exchange.request_done([])
 
-
-    def test_url_err(self):
+    def test_url_unsupported_scheme(self):
         client = HttpClient(loop=self.loop)
         exchange = client.exchange()
         @on(exchange)
@@ -300,10 +320,33 @@ Connection: close
             self.assertEqual(err_msg.__class__, thor.http.error.UrlError)
             self.loop.stop()
 
-        req_uri = b"foo://%s:%i/url_err" % (test_host, test_port)
+        req_uri = b"notascheme://www.example.com:80000/"
         exchange.request_start(b"GET", req_uri, [])
         exchange.request_done([])
 
+    def test_url_nonascii_scheme(self):
+        client = HttpClient(loop=self.loop)
+        exchange = client.exchange()
+        @on(exchange)
+        def error(err_msg):
+            self.assertEqual(err_msg.__class__, thor.http.error.UrlError)
+            self.loop.stop()
+
+        req_uri = "httpü://www.example.com:80000/".encode("utf-8")
+        exchange.request_start(b"GET", req_uri, [])
+        exchange.request_done([])
+
+    def test_url_unbalanced_brackets(self):
+        client = HttpClient(loop=self.loop)
+        exchange = client.exchange()
+        @on(exchange)
+        def error(err_msg):
+            self.assertEqual(err_msg.__class__, thor.http.error.UrlError)
+            self.loop.stop()
+
+        req_uri = "http://[www.example.com:80000/".encode("utf-8")
+        exchange.request_start(b"GET", req_uri, [])
+        exchange.request_done([])
 
     def test_url_port_err(self):
         client = HttpClient(loop=self.loop)
@@ -313,7 +356,7 @@ Connection: close
             self.assertEqual(err_msg.__class__, thor.http.error.UrlError)
             self.loop.stop()
 
-        req_uri = b"http://%s:ABC123/url_port_err" % (test_host)
+        req_uri = b"http://%s:ABC123/url_port_err" % (framework.test_host)
         exchange.request_start(b"GET", req_uri, [])
         exchange.request_done([])
 
@@ -326,10 +369,105 @@ Connection: close
             self.assertEqual(err_msg.__class__, thor.http.error.UrlError)
             self.loop.stop()
 
-        req_uri = b"http://%s:80000/" % (test_host)
+        req_uri = b"http://%s:80000/" % (framework.test_host)
         exchange.request_start(b"GET", req_uri, [])
         exchange.request_done([])
 
+    def test_url_empty_host(self):
+        client = HttpClient(loop=self.loop)
+        exchange = client.exchange()
+        @on(exchange)
+        def error(err_msg):
+            self.assertEqual(err_msg.__class__, thor.http.error.UrlError)
+            self.loop.stop()
+
+        req_uri = b"http://:80000/"
+        exchange.request_start(b"GET", req_uri, [])
+        exchange.request_done([])
+
+    def test_url_dot_host(self):
+        client = HttpClient(loop=self.loop)
+        exchange = client.exchange()
+        @on(exchange)
+        def error(err_msg):
+            self.assertEqual(err_msg.__class__, thor.http.error.UrlError)
+            self.loop.stop()
+
+        req_uri = b"http://.example.com:80000/"
+        exchange.request_start(b"GET", req_uri, [])
+        exchange.request_done([])
+
+    def test_url_nonascii_host(self):
+        client = HttpClient(loop=self.loop)
+        exchange = client.exchange()
+        @on(exchange)
+        def error(err_msg):
+            self.assertEqual(err_msg.__class__, thor.http.error.UrlError)
+            self.loop.stop()
+
+        req_uri = "http://www.ü.example.com:80000/".encode("utf-8")
+        exchange.request_start(b"GET", req_uri, [])
+        exchange.request_done([])
+
+    def test_url_percent_host(self):
+        client = HttpClient(loop=self.loop)
+        exchange = client.exchange()
+        @on(exchange)
+        def error(err_msg):
+            self.assertEqual(err_msg.__class__, thor.http.error.UrlError)
+            self.loop.stop()
+
+        req_uri = b"http://foo%.example.com:80000/"
+        exchange.request_start(b"GET", req_uri, [])
+        exchange.request_done([])
+
+    def test_url_star_host(self):
+        client = HttpClient(loop=self.loop)
+        exchange = client.exchange()
+        @on(exchange)
+        def error(err_msg):
+            self.assertEqual(err_msg.__class__, thor.http.error.UrlError)
+            self.loop.stop()
+
+        req_uri = b"http://example*.com:80000/"
+        exchange.request_start(b"GET", req_uri, [])
+        exchange.request_done([])
+
+    def test_url_startingdigit_label(self):
+        client = HttpClient(loop=self.loop)
+        exchange = client.exchange()
+        @on(exchange)
+        def error(err_msg):
+            self.assertEqual(err_msg.__class__, thor.http.error.UrlError)
+            self.loop.stop()
+
+        req_uri = b"http://www.5test.example.com:80000/"
+        exchange.request_start(b"GET", req_uri, [])
+        exchange.request_done([])
+
+    def test_url_long_label(self):
+        client = HttpClient(loop=self.loop)
+        exchange = client.exchange()
+        @on(exchange)
+        def error(err_msg):
+            self.assertEqual(err_msg.__class__, thor.http.error.UrlError)
+            self.loop.stop()
+
+        req_uri = b"http://%s:80000/" % (b"t" * 256)
+        exchange.request_start(b"GET", req_uri, [])
+        exchange.request_done([])
+
+    def test_url_long_hostname(self):
+        client = HttpClient(loop=self.loop)
+        exchange = client.exchange()
+        @on(exchange)
+        def error(err_msg):
+            self.assertEqual(err_msg.__class__, thor.http.error.UrlError)
+            self.loop.stop()
+
+        req_uri = b"http://wwww.%s.example.com:80000/" % (b"t" * 64)
+        exchange.request_start(b"GET", req_uri, [])
+        exchange.request_done([])
 
     def test_http_version_err(self):
         def client_side(client):
@@ -339,7 +477,7 @@ Connection: close
                 self.assertEqual(err_msg.__class__, thor.http.error.HttpVersionError)
                 self.loop.stop()
 
-            req_uri = b"http://%s:%i/http_version_err" % (test_host, test_port)
+            req_uri = b"http://%s:%i/http_version_err" % (framework.test_host, framework.test_port)
             exchange.request_start(b"GET", req_uri, [])
             exchange.request_done([])
 
@@ -368,7 +506,7 @@ Connection: close
             def response_done(trailers):
                 self.loop.stop()
 
-            req_uri = b"http://%s:%i/http_start_encoding" % (test_host, test_port)
+            req_uri = b"http://%s:%i/http_start_encoding" % (framework.test_host, framework.test_port)
             exchange.request_start(b"GET", req_uri, [])
             exchange.request_done([])
 
@@ -392,7 +530,7 @@ Connection: close
                 self.assertEqual(err_msg.__class__, thor.http.error.HttpVersionError)
                 self.loop.stop()
 
-            req_uri = b"http://%s:%i/protoname_err" % (test_host, test_port)
+            req_uri = b"http://%s:%i/protoname_err" % (framework.test_host, framework.test_port)
             exchange.request_start(b"GET", req_uri, [])
             exchange.request_done([])
 
@@ -421,7 +559,7 @@ Connection: close
                 self.assertEqual(err_msg.__class__, thor.http.error.ConnectError)
                 self.loop.stop()
 
-            req_uri = b"http://%s:%i/close_in_body" % (test_host, test_port)
+            req_uri = b"http://%s:%i/close_in_body" % (framework.test_host, framework.test_port)
             exchange.request_start(b"GET", req_uri, [])
             exchange.request_done([])
 
@@ -440,7 +578,7 @@ Connection: close
     def test_conn_reuse(self):
         self.conn_checked = False
         def client_side(client):
-            req_uri = b"http://%s:%i/conn_reuse" % (test_host, test_port)
+            req_uri = b"http://%s:%i/conn_reuse" % (framework.test_host, framework.test_port)
             exchange1 = client.exchange()
             self.check_exchange(exchange1, {
                 'version': b"1.1",
@@ -502,7 +640,7 @@ Connection: close
     def test_conn_succeed_then_err(self):
         self.conn_checked = False
         def client_side(client):
-            req_uri = b"http://%s:%i/succeed_then_err" % (test_host, test_port)
+            req_uri = b"http://%s:%i/succeed_then_err" % (framework.test_host, framework.test_port)
             exchange1 = client.exchange()
             self.check_exchange(exchange1, {
                 'version': b"1.1",
@@ -552,36 +690,6 @@ Connection: close
         self.assertTrue(self.conn_checked)
 
 
-    def test_HEAD(self):
-        def client_side(client):
-            exchange = client.exchange()
-            self.check_exchange(exchange, {
-                'version': b"1.1",
-                'status': b"200",
-                'phrase': b'OK',
-                'body': b""
-            })
-            @on(exchange)
-            def response_done(trailers):
-                self.loop.stop()
-
-            req_uri = b"http://%s:%i/HEAD" % (test_host, test_port)
-            exchange.request_start(b"HEAD", req_uri, [])
-            exchange.request_done([])
-
-        def server_side(conn):
-            conn.request.send(b"""\
-HTTP/1.1 200 OK
-Content-Type: text/plain
-Content-Length: 5
-Connection: close
-
-""")
-            time.sleep(1)
-            conn.request.close()
-        self.go([server_side], [client_side])
-
-
     def test_req_retry(self):
         def client_side(client):
             exchange = client.exchange()
@@ -595,7 +703,7 @@ Connection: close
             def response_done(trailers):
                 self.loop.stop()
 
-            req_uri = b"http://%s:%i/req_retry" % (test_host, test_port)
+            req_uri = b"http://%s:%i/req_retry" % (framework.test_host, framework.test_port)
             exchange.request_start(b"OPTIONS", req_uri, [])
             exchange.request_done([])
 
@@ -629,7 +737,7 @@ Connection: close
                 self.assertEqual(err_msg.__class__, thor.http.error.ConnectError)
                 self.loop.stop()
 
-            req_uri = b"http://%s:%i/req_retry_fail" % (test_host, test_port)
+            req_uri = b"http://%s:%i/req_retry_fail" % (framework.test_host, framework.test_port)
             exchange.request_start(b"OPTIONS", req_uri, [])
             exchange.request_done([])
 
@@ -662,7 +770,7 @@ Connection: close
             def response_done(trailers):
                 self.loop.stop()
 
-            req_uri = b"http://%s:%i" % (test_host, test_port)
+            req_uri = b"http://%s:%i" % (framework.test_host, framework.test_port)
             exchange.request_start(b"GET", req_uri, [])
             exchange.request_done([])
 
@@ -694,7 +802,7 @@ Connection: close
                 self.assertEqual(err_msg.__class__, thor.http.error.ExtraDataError)
                 self.loop.stop()
 
-            req_uri = b"http://%s:%i" % (test_host, test_port)
+            req_uri = b"http://%s:%i" % (framework.test_host, framework.test_port)
             exchange.request_start(b"GET", req_uri, [])
             exchange.request_done([])
 
@@ -725,7 +833,7 @@ Connection: close
                 self.assertEqual(err_msg.__class__, thor.http.error.ExtraDataError)
                 self.loop.stop()
 
-            req_uri = b"http://%s:%i" % (test_host, test_port)
+            req_uri = b"http://%s:%i" % (framework.test_host, framework.test_port)
             exchange.request_start(b"GET", req_uri, [])
             exchange.request_done([])
 
