@@ -12,7 +12,9 @@ import errno
 import os
 import socket
 import ssl as sys_ssl
+from typing import Union
 
+from thor.dns import lookup
 from thor.loop import LoopBase
 from thor.tcp import TcpClient, TcpConnection
 
@@ -24,6 +26,7 @@ TcpConnection.close_errs.add(sys_ssl.SSL_ERROR_SSL)
 # TODO: TlsServer
 # TODO: expose cipher info, peer info
 
+
 class TlsClient(TcpClient):
     """
     An asynchronous SSL/TLS client.
@@ -44,29 +47,12 @@ class TlsClient(TcpClient):
     conn_handler will be called with the tcp_conn as the argument
     when the connection is made.
     """
+
     def __init__(self, loop: LoopBase = None) -> None:
         TcpClient.__init__(self, loop)
-        try:
-            self.tls_context = sys_ssl.SSLContext(sys_ssl.PROTOCOL_SSLv23)
-        except AttributeError:
-            self.tls_context = None
-
-    def handshake(self) -> None:
-        try:
-            self.sock.do_handshake() # type: ignore
-            self.once('fd_writable', self.handle_connect)
-        except sys_ssl.SSLError as why:
-            if isinstance(why, sys_ssl.SSLWantReadError):
-#            if why == sys_ssl.SSL_ERROR_WANT_READ:
-#                self.once('fd_readable', self.handshake)
-                self.once('fd_writable', self.handshake) # Oh, Linux...
-#            elif why == sys_ssl.SSL_ERROR_WANT_WRITE:
-            elif isinstance(why, sys_ssl.SSLWantWriteError):
-                self.once('fd_writable', self.handshake)
-            else:
-                self.handle_socket_error(why, 'ssl')
-        except socket.error as why:
-            self.handle_socket_error(why, 'ssl')
+        self.tls_context = sys_ssl.create_default_context()
+        self.tls_context.check_hostname = False
+        self.tls_context.verify_mode = sys_ssl.CERT_NONE
 
     # TODO: refactor into tcp.py
     def connect(self, host: bytes, port: int, connect_timeout: float = None) -> None:
@@ -77,54 +63,78 @@ class TlsClient(TcpClient):
         """
         self.host = host
         self.port = port
-        self.once('fd_writable', self.handshake)
-        # FIXME: CAs
-        if self.tls_context:
-            self.sock = self.tls_context.wrap_socket( # type: ignore
-                self.sock,
-                do_handshake_on_connect=False,
-                server_hostname=self.host
+        if connect_timeout:
+            self._timeout_ev = self._loop.schedule(
+                connect_timeout,
+                self.handle_socket_error,
+                socket.error(errno.ETIMEDOUT, os.strerror(errno.ETIMEDOUT)),
             )
-        else: # server_hostname requires 2.7.9
-            self.sock = sys_ssl.wrap_socket(
-                self.sock,
-                cert_reqs=sys_ssl.CERT_NONE,
-                do_handshake_on_connect=False
-            )
-        # TODO: use socket.getaddrinfo(); needs to be non-blocking.
-        try:
-            err = self.sock.connect_ex((host, port))
-        except socket.gaierror as why:
-            self.handle_socket_error(why, 'gai')
+        lookup(host, self._continue_connect)
+
+    def _continue_connect(self, dns_result: Union[str, Exception]) -> None:
+        """
+        Continue connecting after DNS results a result.
+        """
+        if isinstance(dns_result, Exception):
+            self.handle_socket_error(dns_result, "gai")
             return
+        self.once("fd_writable", self.handshake)
+        # FIXME: CAs
+        tcp_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
+        tcp_socket.setblocking(False)
+        self.on("fd_error", self.handle_fd_error)
+        self.register_fd(tcp_socket.fileno(), "fd_writable")
+        self.event_add("fd_error")
+        self.sock = self.tls_context.wrap_socket(  # type: ignore
+            tcp_socket,
+            do_handshake_on_connect=False,
+            server_hostname=self.host.decode("idna"),
+        )
+        try:
+            err = self.sock.connect_ex((dns_result, self.port))
         except socket.error as why:
             self.handle_socket_error(why)
             return
         if err != errno.EINPROGRESS:
             self.handle_socket_error(socket.error(err, os.strerror(err)))
             return
-        if connect_timeout:
-            self._timeout_ev = self._loop.schedule(
-                connect_timeout,
-                self.handle_socket_error,
-                socket.error(errno.ETIMEDOUT, os.strerror(errno.ETIMEDOUT))
-            )
+
+    def handshake(self) -> None:
+        try:
+            self.sock.do_handshake()  # type: ignore
+            self.once("fd_writable", self.handle_connect)
+        except sys_ssl.SSLError as why:
+            if isinstance(why, sys_ssl.SSLWantReadError):
+                #            if why == sys_ssl.SSL_ERROR_WANT_READ:
+                #                self.once('fd_readable', self.handshake)
+                self.once("fd_writable", self.handshake)  # Oh, Linux...
+            #            elif why == sys_ssl.SSL_ERROR_WANT_WRITE:
+            elif isinstance(why, sys_ssl.SSLWantWriteError):
+                self.once("fd_writable", self.handshake)
+            else:
+                self.handle_socket_error(why, "ssl")
+        except socket.error as why:
+            self.handle_socket_error(why, "ssl")
+        except AttributeError as why:
+            # wrap_socket sometimes fails, but the error is caught elsewhere.
+            pass
 
 
 if __name__ == "__main__":
     import sys
     from thor import run
-    test_host = sys.argv[1].encode('utf-8')
+
+    test_host = sys.argv[1].encode("utf-8")
 
     def out(outbytes: bytes) -> None:
-        sys.stdout.write(outbytes.decode('utf-8', 'replace'))
+        sys.stdout.write(outbytes.decode("utf-8", "replace"))
 
     def go(conn: TcpConnection) -> None:
-        conn.on('data', out)
+        conn.on("data", out)
         conn.write(b"GET / HTTP/1.1\r\nHost: %s\r\n\r\n" % test_host)
         conn.pause(False)
 
     c = TlsClient()
-    c.on('connect', go)
+    c.on("connect", go)
     c.connect(test_host, 443)
     run()
